// 
// Module: DummyPixelIterator
// Author: Hassan Siddiqui
//
// Description: 
//
// An NxM pixel iterator, returning a single value for every pixel.
//
// Modifications: See ChangeLog
// 
// $Id: DummyPixelIterator.cc,v 1.1 2012/01/25 15:46:31 rsaxton Exp $

/* General classes */
#include<cassert>

/* SAS-specific classes */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local classes */
#include "DummyPixelIterator.h"

/* Constructors */
DummyPixelIterator::DummyPixelIterator( 
		   		       double xmin,
				       double xmax,
				       unsigned xbins,
				       double ymin,
				       double ymax,
				       unsigned ybins):
  _firstcalled(false),
  _xbins(xbins),
  _wcsconv(0)
{ 
  PROC;
  
  errstr << "DummyPixelIterator called."
	 << Sas::message(Msg::App, Msg::Verbose);

  if (xmin >= xmax)
    {
      errstr << " X range for map ( min =" << xmin
	     << ", max ="<<xmax
	     << " ) is invalid. "
	     << Sas::error("MapXRangeInvalid");

      ALTRETURN;
    }

  if (ymin >= ymax)
    {
      errstr << " Y range for map ( min =" << ymin
	     << ", max ="<<ymax
	     << " ) is invalid. "
	     << Sas::error("MapYRangeInvalid");
    }


  _size = xbins * ybins;

  if (_size == 0)
    {
      errstr << " The number of bins requested is zero "
	     << " ( xbins =" << xbins
	     << ", ybins =" << ybins
	     << Sas::error("MapSizeInvalid");
    }

  double xdelt = (xmax-xmin)/xbins;
  double ydelt = (ymax-ymin)/ybins;


  /* Create WcsConvertor object */
  _wcsconv = new WCSConvertor(
			      "DETX",
			      1,  // 1st pixel in array is [1,1]
			      xmin + (xdelt/2),
			      xdelt,
			      "DETY",
			      1, // 1st pixel in array is [1,1]
			      ymin + (ydelt/2),
			      ydelt);

  ENDPROC;
}


/* Destructor */
DummyPixelIterator::~DummyPixelIterator()
{
  PROC;
  delete _wcsconv;   _wcsconv=0;
  ENDPROC;
}


/* Iterator methods */
void DummyPixelIterator::first()
{
  _firstcalled = true;
  _index=0;
  _xindex=1;
  _yindex=1;
}

void DummyPixelIterator::next()
{
  assert (!isDone());
  _index++;

  _xindex++;
  if (_xindex > _xbins)
    {
      _xindex=1;
      _yindex++;
    }
}

bool DummyPixelIterator::isDone()
{
  PROC;
  assert(_firstcalled);
  RETURN (_index>=_size);
}


PixelElement DummyPixelIterator::currentItem()
{
  /* Convert x and y positions to detector coordinates, 
     using WCS attributes */

  /* [HS] using new WCSConvertor interface */
   CoordPair physcoord = _wcsconv->physicalCoords(_xindex, _yindex);
   PixelElement result (_xindex,
			_yindex,
			physcoord.x(),
			physcoord.y(),
			_wcsconv->xWidth(),
			_wcsconv->yWidth(),
			1);
  RETURN result;
}

unsigned DummyPixelIterator::size() const
{
  PROC;
  RETURN _size;
}

unsigned DummyPixelIterator::xbins() const
{
  PROC;
  RETURN _xbins;
}

/* WCSConvertor object */
const WCSConvertor * DummyPixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}

