//
// Module: SscWcsConvertor
// Author: Richard Saxton 
//
// Description: Class to access the SSCLIB coordinate conversion routines
//
// Modifications: See ChangeLog
//
//
///* General classes */
#include <iostream>
#include <fstream>
//
/* SAS-specific classes */
#include "Dal.h"
#include "CoordPair.h" 
//
///* Local Classes */
#include "SscWcsConvertor.h" 
//#include "WcsCC.h"
//
//

// FORTRAN prototypes
//extern F90MODVARDEF(WcsCC, ssclib_glue, wcs);

F90MODPROCPROTO(void,ssclib_glue, callradectoimpixels)(double &RA, double &DEC,
                         double &xpixel, double &ypixel);
F90MODPROCPROTO(void,ssclib_glue, callradectoskypixels)(double &RA, double &DEC,
                         double &xpixel, double &ypixel);
F90MODPROCPROTO(void,ssclib_glue, callimpixelstoradec)(double &xpixel, 
                                     double &ypixel, double &RA, double &DEC);
F90MODPROCPROTO(void,ssclib_glue, callimpixelstodet)(double &xpixel, 
                                   double &ypixel, double &detx, double &dety);
F90MODPROCPROTO(void, ssclib_glue, setwcs)(unsigned char* filename, size_t *len);
F90MODPROCPROTO(void, ssclib_glue, setwcsref)(unsigned char* filename, size_t *len);
F90MODPROCPROTO(void, ssclib_glue, setatt)(unsigned char* filename, size_t *len);
F90MODPROCPROTO(void,ssclib_glue, calldettoradec)(double &detx, double &dety,
			 double &ra, double &dec);
F90MODPROCPROTO(void,ssclib_glue, callradectodet)(double &ra, double &dec,
                         double &detx, double &dety);
F90MODPROCPROTO(void,ssclib_glue, callxytodet)(double &x, double &y, 
                                                  double &detx, double &dety);
F90MODPROCPROTO(void,ssclib_glue, callrawtodet)(double &rawx, double &rawy, 
                                                  double &detx, double &dety);
F90MODPROCPROTO(void,ssclib_glue, callxytoradec)(double &x, double &y, 
                                                  double &ra, double &dec);
F90MODPROCPROTO(void,ssclib_glue, calldettoxy)(double &detx, double &dety, 
                                                  double &x, double &y);
F90MODPROCPROTO(void,ssclib_glue, calldettotel)(double &detx, double &dety, 
                                                double &theta, double &phi);
F90MODPROCPROTO(void,ssclib_glue, callradectotel)(double &ra, double &dec, 
                                                double &theta, double &phi);
F90MODPROCPROTO(void,ssclib_glue, calldettoraw)(double &detx, double &dety, 
                double &rawx, double &rawy, Integer1 &ccd, bool &isOffChip);

  /* Constructor */
  SscWcsConvertor::SscWcsConvertor() {};

  /* Destructor */
  SscWcsConvertor::~SscWcsConvertor() {};

  /* Method: setWcs() - set up the wcs, within ssclib */
  void SscWcsConvertor::setWcs(const string &imagefile)
  {
    char name[1024];
    sprintf(name, "%s", imagefile.c_str());
    size_t stringLength2=strlen(name);

    unsigned char f90_filename[1024];
    for (unsigned i=0; i<1024; i++)
        f90_filename[i]='\0';

    for (unsigned i=0; i<imagefile.length(); i++)
        f90_filename[i]=(unsigned char)name[i];

    F90MODPROC(ssclib_glue, setwcs)(&f90_filename[0], &stringLength2);
  }

  /* Method: setWcsRef() - set up the wcs, within ssclib */
  void SscWcsConvertor::setWcsRef(const string &imagefile)
  {
    char name[1024];
    sprintf(name, "%s", imagefile.c_str());
    size_t stringLength2=strlen(name);

    unsigned char f90_filename[1024];
    for (unsigned i=0; i<1024; i++)
        f90_filename[i]='\0';
    for (unsigned i=0; i<imagefile.length(); i++)
        f90_filename[i]=(unsigned char)name[i];

    F90MODPROC(ssclib_glue, setwcsref)(&f90_filename[0], &stringLength2);
  }

  /* Method: setAtt - set up the attitude matrix to convert between detector
                      and celestial coordinates, within ssclib */
  void SscWcsConvertor::setAtt(const string &imagefile)
  {
    char name[1024];
    sprintf(name, "%s", imagefile.c_str());
    size_t stringLength2=strlen(name);

    unsigned char f90_filename[1024];
    for (unsigned i=0; i<1024; i++)
        f90_filename[i]='\0';
    for (unsigned i=0; i<imagefile.length(); i++)
        f90_filename[i]=(unsigned char)name[i];

    F90MODPROC(ssclib_glue, setatt)(&f90_filename[0], &stringLength2);
  }

  /* Method: RaDecToImagePixels - converts from ra/dec to image pixel coords */
  CoordPair SscWcsConvertor::raDecToImagePixels(double ra, double dec)
  {
    // Input ra, dec should be in radians
    double xpixel=0, ypixel=0;
   
    // Use the SSCLIB routine raDecToPixel
    F90MODPROC(ssclib_glue, callradectoimpixels)(ra, dec, xpixel, ypixel);

    return CoordPair(xpixel, ypixel);
  }

  /* Method: imagePixelsToRaDec - converts from image pixel coords to ra/dec */
  CoordPair SscWcsConvertor::imagePixelsToRaDec(double xpixel, double ypixel)
  {
    // Output ra, dec will be in radians
    double ra=0.0, dec=0.0;
   
    // Use the SSCLIB routine raDecToPixel
    F90MODPROC(ssclib_glue, callimpixelstoradec)(xpixel, ypixel, ra, dec);

    return CoordPair(ra, dec);
  }

  /* Method: imagePixelsToDet - converts from image pixel to detector coords */
  CoordPair SscWcsConvertor::imagePixelsToDet(double xpixel, double ypixel)
  {
    double detx=0.0, dety=0.0;

    F90MODPROC(ssclib_glue, callimpixelstodet)(xpixel, ypixel, detx, dety);

    return CoordPair(detx, dety);
  }

  /* Method: RaDecToSkyPixels - converts from ra/dec to sky pixel coords */
  CoordPair SscWcsConvertor::raDecToSkyPixels(double ra, double dec)
  {
    // Input ra, dec should be in radians
    double xpixel=0, ypixel=0;
   
    // Use the SSCLIB routine raDecToPixel
    F90MODPROC(ssclib_glue, callradectoskypixels)(ra, dec, xpixel, ypixel);

    return CoordPair(xpixel, ypixel);
  }

  /* Method: RaDecToDet - converts from ra/dec to detector coords */
  CoordPair SscWcsConvertor::raDecToDet(double ra, double dec)
  {
    // Input ra, dec should be in radians
    double detx=0, dety=0;
   
    // Use the SSCLIB routine raDecToPixel
    F90MODPROC(ssclib_glue, callradectodet)(ra, dec, detx, dety);

    return CoordPair(detx, dety);
  }

  /* Method: detToRaDec - converts from detector coords to ra/dec */
  CoordPair SscWcsConvertor::detToRaDec(double detx, double dety)
  {
    double ra, dec;

    F90MODPROC(ssclib_glue, calldettoradec)(detx, dety, ra, dec);

    // NB: ra, dec are returned in radians
    return CoordPair(ra, dec);
  }

  /* Method: detToRaw - converts from detector coords to raw coords */
  CoordPair SscWcsConvertor::detToRaw(double detx, double dety)
  {
    double rawx, rawy;
    Integer1 ccd;
    bool isOffChip; 

    F90MODPROC(ssclib_glue, calldettoraw)(detx, dety, rawx, rawy, 
                                                     ccd, isOffChip);
    return CoordPair(rawx, rawy);
  }

  /* Method: detToCcd - converts from detector coords to CCD number */
  Ccd SscWcsConvertor::detToCcd(double detx, double dety)
  {
    double rawx, rawy;
    Integer1 ccd;
    bool isOffChip; 
    Ccd ccdout;

    F90MODPROC(ssclib_glue, calldettoraw)(detx, dety, rawx, rawy, 
                                                     ccd, isOffChip);
    ccdout=ccd;

    return ccdout;
  }

  /* Method: rawToDet - converts from raw to detector coords */
  CoordPair SscWcsConvertor::rawToDet(double rawx, double rawy)
  {
    double detx, dety;

    F90MODPROC(ssclib_glue, callrawtodet)(rawx, rawy, detx, dety);

    return CoordPair(detx, dety);
  }

  /* Method: xyToDet - converts from sky (X/Y) corrds to detector coords */
  CoordPair SscWcsConvertor::xyToDet(double x, double y)
  {
    double detx, dety;

    F90MODPROC(ssclib_glue, callxytodet)(x, y, detx, dety);

    return CoordPair(detx, dety);
  }

  /* Method: xyToRaDec - converts from sky (X/Y) coords to RA/DEC*/
  CoordPair SscWcsConvertor::xyToRaDec(double x, double y)
  {
    double ra, dec;

    F90MODPROC(ssclib_glue, callxytoradec)(x, y, ra, dec);

    return CoordPair(ra, dec);
  }

  /* Method: detToXy - converts from detector to sky (X/Y) coords */
  CoordPair SscWcsConvertor::detToXy(double detx, double dety)
  {
    double x, y;

    F90MODPROC(ssclib_glue, calldettoxy)(detx, dety, x, y);

    return CoordPair(x, y);
  }

  /* Method: detToTel - converts from detector to tel coords */
  CoordPair SscWcsConvertor::detToTel(double detx, double dety)
  {
    double theta, phi;

    F90MODPROC(ssclib_glue, calldettotel)(detx, dety, theta, phi);

    return CoordPair(theta, phi);
  }

  /* Method: raDecToTel - converts from detector to tel coords */
  CoordPair SscWcsConvertor::raDecToTel(double ra, double dec)
  {
    double theta, phi;

    F90MODPROC(ssclib_glue, callradectotel)(ra, dec, theta, phi);

    return CoordPair(theta, phi);
  }

