// 
// Module: TablePixelIterator
// Author: Hassan Siddiqui
//
// Description: 
//
// Class representation of an iterator that handles DAL table-based
//  detectormaps.
//
// Modifications: See ChangeLog
// 
// $Id: TablePixelIterator.cc,v 1.1 2012/01/25 15:46:31 rsaxton Exp $

// General classes
#include <iostream>
#include<cassert>

// SAS-specific classes
#include "Dal.h"

// General utility functions
#include "errstr.h"
#include "Tracer.h"

// Local classes
#include "TablePixelIterator.h"

/* Constructors */
TablePixelIterator::TablePixelIterator(const string & tablename,
					     const string & xcolname,
					     const string & ycolname,
					     const string & valuecolname) :
  _firstcalled(false),
  _set(0),
  _table(0),
  _xColData(0),
  _yColData(0),
  _valueColData(0),
  _xData(0),
  _yData(0),
  _valueData(0),
  _wcsconv(0)
{
  PROC;
  
  _set = dataSetServer->
    open( tablename, 
	  DataSet::Read );

  _table = _set->table(tablename); 

  _tableelements = _table->rows();
  
  _xColumn     = _table->column(xcolname);
  _yColumn     = _table->column(ycolname);
  _valueColumn = _table->column(valuecolname);

  _xColData     = _xColumn    ->data();
  _yColData     = _yColumn    ->data();
  _valueColData = _valueColumn->data();

  _xData        = _xColData->int16Data();
  _yData        = _yColData->int16Data();
  _valueData    = _valueColData->real64Data();

  /* Determine _xindexsize and _yindexsize from data. 
   * Also check whether data are valid.
   */

  _xindexsize=_xData[0];
  _yindexsize=_yData[0];

  /* Create WcsConvertor object */
  _wcsconv = new WCSConvertor(_xColumn, _yColumn);

  ENDPROC;
}

/* Destructor */
TablePixelIterator::~TablePixelIterator()
{
  PROC;
  
  delete _wcsconv;       _wcsconv=0;
  delete _xColData;      _xColData=0;
  delete _yColData;      _yColData=0;
  delete _valueColData;  _valueColData=0;
 
  dataSetServer->close(_set); _set = 0;
 
  ENDPROC;
}

/* Iterator methods */
void TablePixelIterator::first()
{
  _firstcalled = true;
  _index=0;
}


void TablePixelIterator::next ()
{
  assert (!isDone());
  _index++;
}

bool TablePixelIterator::isDone()
{
  PROC;
  assert(_firstcalled);
  RETURN (_index >= _tableelements);
}


PixelElement TablePixelIterator::currentItem()
{
  PROC;

  assert(_firstcalled);

  /* Convert x and y positions to detector coordinates, 
     using WCS attributes */

  double value = _valueData[_index];

  /* [HS] using new WCSConvertor interface */
  CoordPair physcoord = _wcsconv->physicalCoords(_xData[_index], 
						 _yData[_index]);
  PixelElement result (_xData[_index],
		      _yData[_index],
		       physcoord.x(),
		       physcoord.y(),
		       _wcsconv->xWidth(),
		      _wcsconv->yWidth(),
		      value);

  RETURN result;
}

// Give direct access to the table values without changing coords
double TablePixelIterator::valueAt(unsigned long index)
{
  PROC;
  RETURN _valueData[index];
}

double TablePixelIterator::valueAt()
{
  PROC;
  RETURN _valueData[_index];
}


unsigned TablePixelIterator::size() const
{
  PROC;
  RETURN _tableelements;
}

unsigned TablePixelIterator::xbins() const
{
  PROC;
  
  RETURN _xindexsize;
}

//  int TablePixelIterator::nbinsY() const
//  {
//    PROC;
//    RETURN _yindexsize;
//  }

/* WCSConvertor object */
const WCSConvertor * TablePixelIterator::wcsconvertor() const
{
  PROC;
  RETURN _wcsconv;
}


