// Module: WCSConvertor
//
// Author: Hassan Siddiqui
//
// Description: 
//
// A simple means of reading, converting and
//  propagating WCS information.
// Note that simple linear mapping is performed only.
//
// Modifications: See ChangeLog
// 
// $Id: WCSConvertor.cc,v 1.1 2012/01/25 15:46:31 rsaxton Exp $

/* General */
#include <cmath>

/* SAS-specific */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local classes */
#include "WCSConvertor.h"
#include "NumericalRange.h"

/*** Constructors */

/* Read state from argument list */
WCSConvertor::WCSConvertor( const string & ctype1,
			    int crpix1,
			    double         crval1,
			    double         cdelt1,
			    const string & ctype2,
			    int crpix2,
			    double         crval2,
			    double         cdelt2 ):

  _ctype1(ctype1),
  _crpix1(crpix1),
  _crval1(crval1),
  _cdelt1(cdelt1),

  _ctype2(ctype2),
  _crpix2(crpix2),
  _crval2(crval2),
  _cdelt2(cdelt2)

{
  PROC;
  check();
  ENDPROC;
}

/* Read attribtes from a filename */
WCSConvertor::WCSConvertor(const string & arrayname)
{
  // Will convert this to the SscWcsConvertor soon
  //   // SscWcsConvertor::setWcs(imagefile);
  //
  // In the meantime read the WCS keywords from the top level file 
  DataSet * set = dataSetServer-> open( arrayname, DataSet::Read );
  
  if (
    set->hasAttribute("CTYPE1") &&
    set->hasAttribute("CRPIX1") &&
    set->hasAttribute("CRVAL1") &&
    set->hasAttribute("CDELT1") &&
    set->hasAttribute("CTYPE2") &&
    set->hasAttribute("CRPIX2") &&
    set->hasAttribute("CRVAL2") &&
    set->hasAttribute("CDELT2") )
   {
  
     _ctype1 = set->attribute("CTYPE1")->asString();
  
     _crpix1 = set->attribute("CRPIX1")->asInt();
     _crval1 = set->attribute("CRVAL1")->asReal();
     _cdelt1 = set->attribute("CDELT1")->asReal();
  
     _ctype2 = set->attribute("CTYPE2")->asString();
  
     _crpix2 = set->attribute("CRPIX2")->asInt();
     _crval2 = set->attribute("CRVAL2")->asReal();
     _cdelt2 = set->attribute("CDELT2")->asReal();
  
  /* MegaHack to see if we can get around the DET image in degrees
  *         problem with a rescaling */
     if (_ctype1.find("DET") != string::npos &&
         _ctype1.find("ANG") != string::npos &&
        set->attribute("CUNIT1")->asString().find("deg") != string::npos )
     {
     _cdelt1 = _cdelt1 * 72000.;
     _cdelt2 = _cdelt2 * 72000.;
     _crval1 = _crval1 * 72000.;
     _crval2 = _crval2 * 72000.;
     }
  }
  // Here we will pass these values to the WCS routines but I am 
  // worried about the DETX--ANG construct and how it is handled in ssclib 

  else
    {

      errstr
        << "WCS info for the block name"
        << arrayname
        << " is either non-existent or incomplete. Assuming a 1:1 mapping"
        << Sas::message(Msg::App, Msg::Sparse);

      _ctype1="NULLX";
      _crpix1=0;
      _crval1=0;
      _cdelt1=1;
      _ctype2="NULLY";
      _crpix2=0;
      _crval2=0;
      _cdelt2=1;
    }

}

 
/* Read attributes from an array */
WCSConvertor::WCSConvertor(Array * block)
{
  if ( 
      block->hasAttribute("CTYPE1") &&
      block->hasAttribute("CRPIX1") &&
      block->hasAttribute("CRVAL1") &&
      block->hasAttribute("CDELT1") &&
      block->hasAttribute("CTYPE2") &&
      block->hasAttribute("CRPIX2") &&
      block->hasAttribute("CRVAL2") &&
      block->hasAttribute("CDELT2") )
    {

      _ctype1 = block->attribute("CTYPE1")->asString();
      
      _crpix1 = block->attribute("CRPIX1")->asInt();
      _crval1 = block->attribute("CRVAL1")->asReal();
      _cdelt1 = block->attribute("CDELT1")->asReal();
      
      _ctype2 = block->attribute("CTYPE2")->asString();
      
      _crpix2 = block->attribute("CRPIX2")->asInt();
      _crval2 = block->attribute("CRVAL2")->asReal();
      _cdelt2 = block->attribute("CDELT2")->asReal();

 /* MegaHack to see if we can get around the DET image in degrees
    problem with a rescaling */
      if (_ctype1.find("DET") != string::npos &&
       _ctype1.find("ANG") != string::npos &&
       block->attribute("CUNIT1")->asString().find("deg") != string::npos )
      {
       _cdelt1 = _cdelt1 * 72000.;
       _cdelt2 = _cdelt2 * 72000.;
       _crval1 = _crval1 * 72000.;
       _crval2 = _crval2 * 72000.;
      }
     
    }
  else
    {

      errstr 
	<< "WCS info for the block "
	<< block->qualifiedName()
	<< " is either non-existent or incomplete. Assuming a 1:1 mapping"
	<< Sas::message(Msg::App, Msg::Sparse);
  
      _ctype1="NULLX";
      _crpix1=0;
      _crval1=0;
      _cdelt1=1;
      _ctype2="NULLY";
      _crpix2=0;
      _crval2=0;
      _cdelt2=1;
    }

  check();
}

/* Read attributes from table */
WCSConvertor::WCSConvertor(const Column * xcol, const Column * ycol)
{
  if ( 
      xcol->hasAttribute("TCTYP") &&
      xcol->hasAttribute("TCRPX") &&
      xcol->hasAttribute("TCRVL") &&
      xcol->hasAttribute("TCDLT") &&
      ycol->hasAttribute("TCTYP") &&
      ycol->hasAttribute("TCRPX") &&
      ycol->hasAttribute("TCRVL") &&
      ycol->hasAttribute("TCDLT") )
    {

      _ctype1 = xcol->attribute("TCTYP")->asString();
      _crpix1 = xcol->attribute("TCRPX")->asInt();
      _crval1 = xcol->attribute("TCRVL")->asReal();
      _cdelt1 = xcol->attribute("TCDLT")->asReal();
      
      _ctype2 = ycol->attribute("TCTYP")->asString();
      _crpix2 = ycol->attribute("TCRPX")->asInt();
      _crval2 = ycol->attribute("TCRVL")->asReal();
      _cdelt2 = ycol->attribute("TCDLT")->asReal();
    }
  else
    {
      errstr 
	<< "WCS info for the columns "
	<< xcol->qualifiedName()
	<< " and "
	<< xcol->qualifiedName()
	<< " is either non-existent or incomplete. Assuming a 1:1 mapping"
	<< Sas::message(Msg::App, Msg::Sparse);

      _ctype1="NULLX";
      _crpix1=0;
      _crval1=0;
      _cdelt1=1;
      _ctype2="NULLY";
      _crpix2=0;
      _crval2=0;
      _cdelt2=1;
      
    }

  check();
}

/* Use 1:1 mapping for two input axis types */
WCSConvertor::WCSConvertor(const string & xaxis, const string & yaxis):
  _ctype1(xaxis),
  _crpix1(0),
  _crval1(0),
  _cdelt1(1),
  _ctype2(yaxis),
  _crpix2(0),
  _crval2(0),
  _cdelt2(1)
{
  PROC;
  check();
  ENDPROC;
};

/* Default constructor */
WCSConvertor::WCSConvertor():
  _ctype1("NULLX"),
  _crpix1(0),
  _crval1(0),
  _cdelt1(1),
  _ctype2("NULLY"),
  _crpix2(0),
  _crval2(0),
  _cdelt2(1)
{
  PROC;
  ENDPROC;
};

/* Destructors */
WCSConvertor::~WCSConvertor(){};

bool WCSConvertor::check() const
{
  PROC;

  errstr << "WCSConvertor "
	 << " _ctype1=" << _ctype1
	 << " _crpix1=" << _crpix1
	 << " _crval1=" << _crval1
	 << " _cdelt1=" << _cdelt1
	 << " _ctype2=" << _ctype2
	 << " _crpix2=" << _crpix2
	 << " _crval2=" << _crval2
	 << " _cdelt2=" << _cdelt2
	 << Sas::message(Msg::App, Msg::Noisy); 

  if (_ctype1=="RA---TAN")
    {
      if (_ctype2=="DEC--TAN")
	{
	  ALTRETURNV( true );
	}

      errstr 
	<< "The axis types " << _ctype1 
	<< " and " << _ctype2 << " are not compatible."
	<< Sas::error("InvalidWCSType");

      ALTRETURNV( false );
    }

  RETURN true;
}
  
/* Accessors */

/* Method: physicalCoords - returns the physical coordinates of the
   input position. For detector coordinates the input is a pixel position
   on an image, for RA/DEC output the input values are X/Y sky coordinates */

CoordPair WCSConvertor::physicalCoords(double xpix, double ypix) const
{
  PROC;
  /* Compute relative coordinates in physical units
     (section 2.1 of      
     'Representations of Celestial Coordinates in FITS',
     Eric Greisen and Mark Calabretta (1994).
  */

  /* Correct positions to pixel centres and compute
     position relative to reference point */
  //double rx = xpix + 0.5 - _crpix1;
  //double ry = ypix + 0.5 - _crpix2;
  double rx = xpix - _crpix1;
  double ry = ypix - _crpix2;

  double lx = rx * _cdelt1;
  double ly = ry * _cdelt2;

  errstr 
    << "WCSConvertor: From input pixel position of ["
    << xpix << "," << ypix << "] got relative position of ["
    << rx << "," << ry << "] and a linearly mapped position of ["
    << lx << "," << ly << "]"
    << Sas::message(Msg::App, Msg::Noisy);

  if (_ctype1 != "RA---TAN")   
    {

      double px = lx + _crval1;
      double py = ly + _crval2;

      errstr 
	<< "WCSConvertor: Returning linearly-mapped PHYSICAL coords of ["
	<< px
	<< "," 
	<< py << "]"
	<< Sas::message(Msg::App, Msg::Noisy);

      /* Linear mapping */
      ALTRETURNV( CoordPair(px, py) );
    }
  
  /* Non-linear mapping from image to equatorial coords 
     Following section 6 and equations 132, 10, 15 and 11 of
     Greisen and Calabretta (1994).
   */

  // Pre-compute values
  double pi180 = 180*M_1_PI;

  double phi   = atan2(lx, -1.0 * ly);
  double r     = sqrt(lx*lx + ly*ly);

  /* Avoid division by zero */
  double theta = (r<=FLT64_MIN) ? M_PI_2 : atan (pi180/r);
  errstr 
    << "WCSConvertor: "
    << "r=" << r
    << " theta=" << theta*pi180
    << " phi=" << phi*pi180
    << Sas::message(Msg::App, Msg::Noisy);

  double decrefrad = _crval2/pi180;
  double deltarad  = 
    asin( sin(theta)*sin(decrefrad) - 
	 cos(theta)*cos(phi)*cos(decrefrad) );

  double dalpharad = asin(cos(theta)*sin(phi)/cos(deltarad));
 
  double alpha = dalpharad*pi180 + _crval1;
  double delta = deltarad*pi180;

  errstr 
    << "WCSConvertor: returning ra=" << alpha
    << " dec=" << delta
    << " dalpharad=" << dalpharad
    << " crval1=" << _crval1
    << Sas::message(Msg::App, Msg::Noisy);

  RETURN CoordPair(alpha, delta);
}

/* Method: linearCoords - returns the linear coordinates of the
   input position. For detector coordinates the input is a pixel position
   on an image, for input values of X/Y sky coordinates the output is
   the sky position on the tangent plane, i.e. not the RA/DEC. */

/* Return the linear coordinates from the tangent plane */
CoordPair WCSConvertor::linearCoords(double xpix, double ypix) const
{
  PROC;
  /* Correct positions to pixel centres and compute
     position relative to reference point */
  double rx = xpix - _crpix1;
  double ry = ypix - _crpix2;

  double lx = rx * _cdelt1;
  double ly = ry * _cdelt2;

  double px = lx + _crval1;
  double py = ly + _crval2;

  errstr 
	<< "WCSConvertor: Returning linear coords of ["
	<< px
	<< "," 
	<< py << "]"
	<< Sas::message(Msg::App, Msg::Noisy);

  RETURN CoordPair(px, py);
}

double WCSConvertor::xWidth() const
{
  PROC;
  RETURN _cdelt1;
}

double WCSConvertor::yWidth() const
{
  PROC;
  RETURN _cdelt2;
}


string WCSConvertor::xType() const
{
  PROC;
  RETURN _ctype1;
}

string WCSConvertor::yType() const
{
  PROC;
  RETURN _ctype2;
}

/* Method physicalRefCoords - returns the coords at the reference point */
CoordPair WCSConvertor::physicalRefCoords() const
{
  PROC;
  RETURN CoordPair (_crval1, _crval2);
}
 
/* Method RefPixel - returns the pixel position of the reference point */
CoordPair WCSConvertor::physicalRefPixel() const
{
  PROC;
  RETURN CoordPair (_crpix1, _crpix2);
}
 

/* Add attributes to an array */
void WCSConvertor::addAttributesTo(Array * block) const
{
  *(block->addAttribute("CTYPE1")) = _ctype1;
  *(block->addAttribute("CRPIX1")) = (int) _crpix1;
  *(block->addAttribute("CRVAL1")) = _crval1;
  *(block->addAttribute("CDELT1")) = _cdelt1;

  *(block->addAttribute("CTYPE2")) = _ctype2;
  *(block->addAttribute("CRPIX2")) = (int) _crpix2;
  *(block->addAttribute("CRVAL2")) = _crval2;
  *(block->addAttribute("CDELT2")) = _cdelt2;

}

/* Add attributes to a table */
void WCSConvertor::addAttributesTo(Column * xcol, Column * ycol) const
{
  *(xcol->addAttribute("TCTYP")) = _ctype1;
  *(xcol->addAttribute("TCRPX")) = (int) _crpix1;
  *(xcol->addAttribute("TCRVL")) = _crval1;
  *(xcol->addAttribute("TCDLT")) = _cdelt1;
  
  *(ycol->addAttribute("TCTYP")) = _ctype2;
  *(ycol->addAttribute("TCRPX")) = (int) _crpix2;
  *(ycol->addAttribute("TCRVL")) = _crval2;
  *(ycol->addAttribute("TCDLT")) = _cdelt2;
}
